# Copyright (c) Shipt, Inc.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import TypeVar

from volley.models.pydantic_model import GenericMessage, QueueMessage

GenericMessageType = TypeVar("GenericMessageType", bound=GenericMessage)

__all__ = ["GenericMessage", "QueueMessage", "GenericMessageType"]



import signal
from typing import Any

from volley.logging import logger


class GracefulKiller:
    def __init__(self) -> None:
        self.kill_now = False
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self, signum: int, frame: Any) -> None:  # pylint: disable=W0613
        logger.warning("Received kill: %s", signum)
        self.kill_now = True



from typing import Any

import msgpack

from volley.serializers.base import BaseSerialization


class MsgPackSerialization(BaseSerialization):
    def serialize(self, message: dict[Any, Any]) -> bytes:
        serialized: bytes = msgpack.packb(message)
        return serialized

    def deserialize(self, message: bytes) -> dict[str, Any]:
        deserialized: dict[str, Any] = msgpack.unpackb(message)
        return deserialized



# Copyright (c) Shipt, Inc.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from abc import ABC, abstractmethod
from time import time
from typing import Any, Optional, Tuple

from prometheus_client import Summary

from volley.logging import logger
from volley.serializers.base import BaseSerialization

PROCESS_TIME = Summary("data_model_process_seconds", "Time handling message to data model (seconds)", ["process"])


class BaseModelHandler(ABC):
    """Base definition for a schema validator

    Provides the method for constructing an object with a schema
    """

    @abstractmethod
    def construct(self, message: Any, schema: Any) -> Any:
        """turns a raw message into a data model

        The purpose is to provide the definitions on how to turn a
        message from a serializer into a message that an application
        is ready to consume. If the serializer will return a dict, then
        the type of the param `message` would be `dict`. The construct
        method would parse the `dict` into whatever data model that the
        application is expecting to received.

        Args:
            message (Any): A message to construct into a data model specified by `schema`.
                The message can be of any type so long as  .construct() and .deconstruct()
                have implementation details to handle the type.
            schema (Any): The data model definition. `message` is used to create an instance
                of `schema`

        Returns:
            Any: an instance of class `schema`
        """

    @abstractmethod
    def deconstruct(self, model: Any) -> Any:
        """turns data model into a raw type"""


def message_model_handler(
    message: Any,
    schema: Any,
    model_handler: Optional[BaseModelHandler] = None,
    serializer: Optional[BaseSerialization] = None,
) -> Tuple[Any, bool]:
    """handles converting data from connector to data model for application

    Args:
        message (Any): [description]
        schema (Any): [description]
        model_handler (Optional[BaseModelHandler], optional): [description]. Defaults to None.
        serializer (Optional[BaseSerialization], optional): [description]. Defaults to None.

    Returns:
        Tuple[Any, bool]: (data_model, True) on success, or (message, False)

    Raises:
        Exception: surfaced serializer or model handler
    """
    # input serialization
    # optional - some frameworks may handle model constuction and serialization in one-shot
    deserialized_msg: Any
    if serializer is not None:
        try:
            start = time()
            deserialized_msg = serializer.deserialize(message)
            duration = time() - start
            PROCESS_TIME.labels("deserialize").observe(duration)

        except Exception:
            logger.exception("Deserialization failed message=%s - serializer=%s", message, serializer)
            return (message, False)
    else:
        # serializer is disabled
        deserialized_msg = message

    # model construction
    if model_handler is not None:
        try:
            start = time()
            data_model = model_handler.construct(
                message=deserialized_msg,
                schema=schema,
            )
            duration = time() - start
            PROCESS_TIME.labels("construct").observe(duration)
            return (data_model, True)
        except Exception:
            logger.exception("Failed model construction. message=%s - schema=%s", deserialized_msg, schema)
            return (message, False)
    else:
        return (deserialized_msg, True)


def model_message_handler(
    data_model: Any,
    model_handler: Optional[BaseModelHandler] = None,
    serializer: Optional[BaseSerialization] = None,
) -> Any:
    """handles the convertion of a data model to data ready for a connector
    Most commonly this is a Pydantic model to bytes
    """

    try:
        # one block - if any of this fails, we crash hard because
        # we dont want to try to recover if we cant publish

        # convert data model to raw type
        start = time()
        if model_handler is not None:
            raw = model_handler.deconstruct(data_model)
        else:
            raw = data_model
        duration = time() - start
        PROCESS_TIME.labels("deconstruct").observe(duration)

        if serializer is not None:
            start = time()
            # serialize if asked
            serialized = serializer.serialize(raw)
            duration = time() - start
            PROCESS_TIME.labels("serialize").observe(duration)
            return serialized
        else:
            return raw
    except Exception:
        logger.exception("failed transporting message to connector: data_model=%s", data_model)
        raise



from volley.connectors.confluent import ConfluentKafkaConsumer, ConfluentKafkaProducer
from volley.connectors.rsmq import RSMQConsumer, RSMQProducer

__all__ = [
    "ConfluentKafkaConsumer",
    "ConfluentKafkaProducer",
    "RSMQConsumer",
    "RSMQProducer",
]



# Copyright (c) Shipt, Inc.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import importlib
from pathlib import Path
from typing import Any, Dict, Union

from yaml import Loader, load

GLOBALS = Path(__file__).parent.resolve().joinpath("global.yml")


def load_yaml(file_path: Union[str, Path]) -> Dict[str, Any]:
    """loads a yaml to dict from Path object
    Raises FileNotFoundError
    """
    if isinstance(file_path, str):
        path = Path(file_path)
    else:
        path = file_path
    with path.open() as f:
        cfg: Dict[str, Any] = load(f, Loader=Loader)
    return cfg


def import_module_from_string(module_str: str) -> type:
    """returns the module given its string path
    for example:
        'volley.data_models.GenericMessage'
    is equivalent to:
        from volley.data_models import GenericMessage
    """
    modules = module_str.split(".")
    class_obj = modules[-1]
    pathmodule = ".".join(modules[:-1])
    module = importlib.import_module(pathmodule)
    t: type = getattr(module, class_obj)
    return t


def get_configs() -> Dict[str, Dict[str, Any]]:
    return load_yaml(GLOBALS)



import os
import time
from dataclasses import dataclass
from typing import Any, Optional, Union

from prometheus_client import Summary
from rsmq import RedisSMQ
from tenacity import retry, stop_after_attempt, wait_exponential, wait_fixed

from volley.connectors.base import BaseConsumer, BaseProducer
from volley.data_models import QueueMessage
from volley.logging import logger

PROCESS_TIME = Summary("redis_process_time_seconds", "Time spent interacting with rsmq", ["operation"])


class RSMQConfigError(Exception):
    """problems with RSMQ config"""


@dataclass
class RSMQConsumer(BaseConsumer):
    """Handles consuming messages from Redis Simple Message Queue
    [https://github.com/mlasevich/PyRSMQ#quick-intro-to-rsmq](https://github.com/mlasevich/PyRSMQ#quick-intro-to-rsmq)
    - pseudo example usage and configurations available on volley.Engine init:
    ```python hl_lines="4 6-12"
    from volley import Engine
    config = {
        "my_rsmq_queue": {
            "consumer": "volley.connectors.RSMQConsumer",
            "serializer": "volley.serializers.OrJsonSerialization",
            "config": {
                "host": "<hostname for the redis instance>",
                "port": "port for the redis instance, defaults to 6379",
                "vt": 120,
                "options": {
                    "decode_responses": False,
                }
            }
        }
    }
    app = Engine(
        input_queue="my_rsmq_queue",
        queue_config=config,
        ...
    )
    ```
    """

    def __post_init__(self) -> None:
        if "host" in self.config:
            # pass the value directly to the constructor
            pass
        elif host := os.getenv("REDIS_HOST"):
            self.config["host"] = host
        else:
            raise RSMQConfigError("RSMQ host not found in environment nor config")

        if "options" in self.config:
            if "decode_responses" in self.config["options"]:
                # if client is providing their own value, use it
                pass
            else:
                # otherwise set it to False. Serialization is handled elsewhere
                self.config["options"]["decode_responses"] = False
        else:
            # if no options provided, then provide these default options
            self.config["options"] = {"decode_responses": False}

        defaults = {"qname": self.queue_name, "exceptions": True, "vt": 60}

        defaults.update(self.config)
        self.config = defaults
        logger.info("RSMQ Consumer configs %s", self.config)
        self.queue = RedisSMQ(**self.config)
        logger.info("Creating queue: %s", self.queue_name)
        self.queue.createQueue().exceptions(False).execute()

    def consume(self) -> Optional[QueueMessage]:
        """Polls RSMQ for a single message.

            RSMQ consume returns a dictionary from the queue:
            # https://github.com/mlasevich/PyRSMQ/blob/master/README.md
            {
                message: Any - the message content
                rc: int - receive count - how many times this message was received
                ts: int - unix timestamp of when the message was originally sent
                id: str - message id
            }

        Returns:
            Optional[QueueMessage]: The message body and it's RSMQ object.
        """
        _start = time.time()
        msg = (
            self.queue.receiveMessage(qname=self.queue_name, quiet=True, vt=self.config["vt"])
            .exceptions(False)
            .execute()
        )
        _duration = time.time() - _start
        PROCESS_TIME.labels("read").observe(_duration)
        if isinstance(msg, dict):
            if isinstance(msg["id"], bytes):
                # message id is bytes when decode_response = False
                msg["id"] = msg["id"].decode("utf-8")
            return QueueMessage(message_context=msg["id"], message=msg["message"])
        else:
            return None

    def on_success(self, message_context: str) -> None:
        _start = time.time()
        self.delete_message(message_id=message_context)
        _duration = time.time() - _start
        PROCESS_TIME.labels("delete").observe(_duration)

    def on_fail(self, message_context: str) -> None:
        """message will become visible once visibility timeout expires"""
        # self.queue.changeMessageVisibility(qname=queue_name, id=message_context, vt=0)
        logger.error(
            "Failed producing message id %s." "Message will reappear in queue `%s` when timeout expires",
            message_context,
            self.queue_name,
        )

    def shutdown(self) -> None:
        self.queue.quit()

    @retry(reraise=True, wait=wait_exponential(multiplier=1, min=4, max=10))
    def delete_message(self, message_id: str) -> bool:
        """wrapper function to handle retries
        retrying forever with exponential backoff
        """
        result: bool = self.queue.deleteMessage(qname=self.queue_name, id=message_id).execute()
        if result:
            return result
        else:
            err = f"Failed deleting message: '{message_id}' from queue: '{self.queue_name}'"
            logger.critical(err)
            raise TimeoutError(err)


@dataclass
class RSMQProducer(BaseProducer):
    def __post_init__(self) -> None:
        # delivery reports are synchronous
        self.callback_delivery = False

        if "host" in self.config:
            # pass the value directly to the constructor
            pass
        elif host := os.getenv("REDIS_HOST"):
            self.config["host"] = host
        else:
            raise RSMQConfigError("RSMQ host not found in environment nor config")

        defaults = {
            "qname": self.queue_name,
            "delay": 0,
            "exceptions": True,
        }

        defaults.update(self.config)
        self.config = defaults
        logger.info("RSMQ Producer configs: %s", self.config)
        self.queue = RedisSMQ(**self.config)
        logger.info("Creating queue: %s", self.queue_name)
        self.queue.createQueue().exceptions(False).execute()

    def produce(
        self, queue_name: str, message: bytes, message_context: Optional[Any] = None, **kwargs: Union[str, int]
    ) -> bool:
        _start = time.time()
        status: bool = self.send_message(queue_name=queue_name, message=message, produce_cfg=kwargs)
        _duration = time.time() - _start
        PROCESS_TIME.labels("write").observe(_duration)
        return status

    @retry(reraise=True, stop=stop_after_attempt(5), wait=wait_fixed(2))
    def send_message(self, queue_name: str, message: bytes, produce_cfg: dict[str, Any]) -> bool:
        """wrapper function to handle retries retrying"""
        msg_id: str = self.queue.sendMessage(
            qname=queue_name,
            message=message,
            encode=produce_cfg.get("encode", False),
            delay=produce_cfg.get("delay", None),
            quiet=produce_cfg.get("quiet", False),
        ).execute()
        return bool(msg_id)

    def shutdown(self) -> None:
        self.queue.quit()



import asyncio
import contextvars
import functools
from typing import Any, Awaitable, Callable


async def run_in_threadpool(func: Callable[..., Any], *args: Any) -> Any:
    loop = asyncio.get_event_loop()
    child = functools.partial(func, *args)
    context = contextvars.copy_context()
    func = context.run
    args = (child,)
    return await loop.run_in_executor(None, func, *args)


async def run_worker_function(func: Any, message: Any, is_coroutine: bool) -> Any:
    if is_coroutine:
        return await func(message)
    else:
        return await run_in_threadpool(func, message)


def run_async(func: Callable[..., Awaitable[Any]]) -> Callable[..., None]:
    def wrapper() -> None:
        asyncio.run(func())

    return wrapper



# Copyright (c) Shipt, Inc.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from abc import ABC, abstractmethod
from typing import Any


class BaseSerialization(ABC):
    """Base class for serializing and deserializing queue data"""

    @abstractmethod
    def serialize(self, message: dict[Any, Any]) -> bytes:
        """serialize to queue"""

    @abstractmethod
    def deserialize(self, message: bytes) -> Any:
        """deserialize from queue"""



# Copyright (c) Shipt, Inc.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from copy import deepcopy
from enum import Enum, auto
from typing import Any, Dict, Optional

from pydantic import BaseModel, root_validator, validator

from volley.config import get_configs
from volley.logging import logger


class ConnectionType(Enum):
    """types of connections to queues"""

    PRODUCER: auto = auto()
    CONSUMER: auto = auto()


class Profile(BaseModel):

    connection_type: ConnectionType
    # all str - path to the objects
    consumer: Optional[str]
    producer: Optional[str]
    model_handler: Optional[str]
    data_model: Optional[str]
    serializer: Optional[str]

    @validator("model_handler", "data_model", "serializer")
    @classmethod
    def validate_nullable(cls, value: Optional[str]) -> Optional[str]:
        """Ensures str or None types parse as `None`
        This needs to be the top (first) validator
        """
        if str(value).lower() not in ["none", "disabled"]:
            return value
        else:
            return None

    @root_validator
    @classmethod
    def validate_connectors(cls, values: dict[str, Any]) -> Any:
        if values.get("producer") is None and values["connection_type"] == ConnectionType.PRODUCER:
            raise ValueError("Invalid Profile. Must provide a producer for output queues.")
        if values.get("consumer") is None and values["connection_type"] == ConnectionType.CONSUMER:
            raise ValueError("Invalid Profile. Must provide a consumer for input queues.")
        return values

    @root_validator
    @classmethod
    def validate_handler(cls, values: dict[str, Any]) -> Any:
        """Volley cannot construct data into a data model without a model handler
        both model_handler and data_model can be None, however
        """
        data_model = values.get("data_model")
        model_handler = values.get("model_handler")
        if data_model is not None and model_handler is not None:
            # happy path - both are provided
            return values
        elif data_model is None and model_handler is None:
            # happy, but unconvential path
            logger.info("model_handler and data_model are disabled")
            return values
        else:
            raise ValueError(
                "Invalid Profile. Must provide both or none of model_handler|data_model"
                f"{data_model=} -- {model_handler=}"
            )


def construct_profiles(queue_configs: dict[str, dict[str, Any]]) -> Dict[str, Profile]:
    """Constructs a profile from each queue configuration provided by user
    User behavior:
        1. Specifies a named profile with no overrides
        2. Specifies a named profile and provide overrides
        3. Does not specify a profile and provides all values
    """
    supported_profiles: Dict[str, Dict[str, str]] = get_configs()["profiles"]

    constructed_profiles: Dict[str, Profile] = {}

    for q_name, q_cfg in queue_configs.items():
        profile_requested = q_cfg.get("profile")
        if profile_requested in supported_profiles:
            this_profile: Dict[str, str] = deepcopy(supported_profiles[profile_requested])
            # override the requested profile with any user provided cfgs
            this_profile.update(q_cfg)
        elif profile_requested is None:
            # user can provide all required configs
            # and is validated in Profile()
            this_profile = deepcopy(q_cfg)
        else:
            raise ValueError(
                f"`{profile_requested}` is not a valid profile name. "
                f"Available profiles: `{list(supported_profiles.keys())}`"
            )
        constructed_profiles[q_name] = Profile(**this_profile)

    return constructed_profiles



import os
from dataclasses import dataclass
from threading import Thread
from typing import Any, Dict, Optional, Union

from confluent_kafka import Consumer, Message, Producer
from prometheus_client import Counter

from volley.connectors.base import BaseConsumer, BaseProducer
from volley.data_models import QueueMessage
from volley.logging import logger

DELIVERY_STATUS = Counter("delivery_report_status", "Kafka delivered message", ["status"])


@dataclass
class ConfluentKafkaConsumer(BaseConsumer):
    """
    Use for consuming a single message from a single Kafka topic.

    Built on confluent-kafka-python/librdkafka. Offsets are stored in librdkafka
        and commited according to auto.commit.interval.ms
        https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md
    """

    poll_interval: float = 10
    auto_offset_reset: str = "earliest"
    auto_commit_interval_ms: int = 3000

    def __post_init__(self) -> None:  # noqa: C901
        # mapping of partition/offset of the last stored commit
        self.last_offset: dict[int, int] = {}

        self.config = handle_creds(self.config)

        if "auto.offset.reset" not in self.config:
            logger.info("Assigning auto.offset.reset default: %s", self.auto_offset_reset)
            self.config["auto.offset.reset"] = self.auto_offset_reset

        if "auto.commit.interval.ms" not in self.config:
            logger.info("Assigning auto.commit.interval.ms default: %s", self.auto_commit_interval_ms)
            self.config["auto.commit.interval.ms"] = self.auto_commit_interval_ms

        # self.config provided from base Consumer class
        # consumer group assignment
        # try config, then env var, then command line argument w/ env
        if "group.id" in self.config:
            # we'll pass the config directly into Kafka constructor
            pass
        else:
            try:
                self.config["group.id"] = os.environ["KAFKA_CONSUMER_GROUP"]
            except KeyError:
                logger.exception("KAFKA_CONSUMER_GROUP not specified in env variable or group.id in config dict")
                raise

        if "poll_interval" in self.config:
            # poll_interval can be overridden by a user provided Engine init config
            # but it should not be passed to base Confluent Consumer init
            # confluent will ignore it with a warning, however
            self.poll_interval = self.config.pop("poll_interval")

        # this is not overridable, due to the behavior in self.on_success()
        # on_success calls store_offsets()
        self.config["enable.auto.offset.store"] = False
        self.c = Consumer(self.config, logger=logger)
        logger.info("Kafka Consumer Configuration: %s", self.config)
        self.c.subscribe([self.queue_name])
        logger.info("Subscribed to %s", self.queue_name)

    def consume(  # type: ignore
        self,
    ) -> Optional[QueueMessage]:
        message = self.c.poll(self.poll_interval)
        if message is None:
            pass
        elif message.error():
            logger.warning(message.error())
            message = None
        else:
            return QueueMessage(message_context=message, message=message.value())

    def on_success(self, message_context: Message) -> None:
        """stores any offsets that are able to be stored"""
        partition = message_context.partition()
        this_offset = message_context.offset()

        # see if we've already committed this or a higher offset
        # delivery report from kafka producer are not guaranteed to be in order
        # that they were produced
        # https://github.com/confluentinc/confluent-kafka-python/issues/300#issuecomment-358416432
        try:
            last_commit = self.last_offset[partition]
        except KeyError:
            # first message from this partition
            self.last_offset[partition] = this_offset
            self.c.store_offsets(message_context)
            return
        if this_offset > last_commit:
            self.c.store_offsets(message_context)  # committed according to auto.commit.interval.ms
            self.last_offset[partition] = this_offset

    def on_fail(self, message_context: Message) -> None:
        logger.info(
            "Downstream failure. Did not commit offset: %d, partition: %d, message: %s",
            message_context.offset(),
            message_context.partition(),
            message_context.value(),
        )

    def shutdown(self) -> None:
        self.c.close()
        logger.info("Successfully commit offsets and left consumer group %s", self.config.get("group.id"))


@dataclass
class ConfluentKafkaProducer(BaseProducer):
    compression_type: str = "gzip"
    thread: bool = False
    poll_thread_timeout: float = 1.0

    def __post_init__(self) -> None:  # noqa: C901
        self.callback_delivery = True
        self.config = handle_creds(self.config)
        if "compression.type" not in self.config:
            logger.info("Assigning compression.type default: %s", self.compression_type)
            self.config["compression.type"] = self.compression_type

        self.p = Producer(self.config, logger=logger)
        # self.config comes from super class
        logger.info("Kafka Producer Configuration: %s", self.config)

        # producer poll thread
        self.kill_poll_thread = False
        if "poll_thread_timeout" in self.config:
            self.poll_thread_timeout = self.config.pop("poll_thread_timeout")

    def init_callbacks(self, consumer: BaseConsumer, thread: bool = True) -> None:
        # daemon thread - this can be aggressively killed on shutdown
        # Volley will call shutdown() which will trigger confluent.Producer.flush()
        # flush() will trigger acked() and call consumers on_success/on_fail
        self.on_success = consumer.on_success
        self.on_fail = consumer.on_fail
        if thread:
            self.thread = True
            self.poll_thread = Thread(target=self.handle_poll, daemon=True)
            self.poll_thread.start()

    def handle_poll(self) -> None:
        while not self.kill_poll_thread:
            self.p.poll(0.2)

    def acked(self, err: Optional[str], msg: Message, consumer_context: Any) -> None:
        if err is not None:
            logger.critical(
                "Failed delivery to %s, message: %s, error: %s",
                msg.topic(),
                msg.value(),
                err,
            )
            DELIVERY_STATUS.labels("fail").inc()
            self.on_fail(consumer_context)  # type: ignore

        else:
            logger.info(
                "Successful delivery to %s, partion: %d, offset: %d", msg.topic(), msg.partition(), msg.offset()
            )
            DELIVERY_STATUS.labels("success").inc()
            self.on_success(consumer_context)  # type: ignore

    def produce(self, queue_name: str, message: bytes, message_context: Any, **kwargs: Union[str, int]) -> bool:
        self.p.produce(
            key=kwargs.get("key"),
            topic=queue_name,
            value=message,
            headers=kwargs.get("headers"),
            callback=lambda err, msg, obj=message_context: self.acked(err, msg, obj),
        ),

        logger.info("Sent to topic: %s", queue_name)
        return True

    def shutdown(self) -> None:
        self.p.flush()
        if self.thread:
            self.kill_poll_thread = True
            self.poll_thread.join(self.poll_thread_timeout)


def handle_creds(config_dict: Dict[str, Any]) -> Dict[str, Any]:
    if "bootstrap.servers" in config_dict:
        pass
    else:
        try:
            config_dict["bootstrap.servers"] = os.environ["KAFKA_BROKERS"]
        except KeyError:
            logger.exception("Kafka brokers not specified, set in config dict or env var KAFKA_BROKERS")
            raise

    # No key == dev mode
    if config_dict.get("sasl.username") and config_dict.get("sasl.password"):
        pass
    else:
        sasl_username = os.getenv("KAFKA_KEY")
        sasl_password = os.getenv("KAFKA_SECRET")
        if (sasl_username is not None) and (sasl_password is not None):
            logger.info(
                "KAFKA_KEY and KAFKA_SECRET found in environment. Assigning security.protocol and sasl.mechanism"
            )
            config_dict["sasl.username"] = sasl_username
            config_dict["sasl.password"] = sasl_password
            config_dict["security.protocol"] = "SASL_SSL"
            config_dict["sasl.mechanism"] = "PLAIN"
    return config_dict



from typing import Any, Dict, Type, TypeVar

from pydantic import BaseModel, Extra

from volley.models.base import BaseModelHandler

BaseModelType = TypeVar("BaseModelType", bound=BaseModel)


class PydanticModelHandler(BaseModelHandler):
    """for pydantic model that offloads serialization to serializer"""

    def construct(self, message: dict[str, Any], schema: Type[BaseModelType]) -> BaseModelType:
        """coverts a dict to a Pydantic model"""
        return schema.parse_obj(message)

    def deconstruct(self, model: BaseModelType) -> Dict[str, Any]:
        """converts a pydantic model to a dict"""
        return model.dict()


class PydanticParserModelHandler(BaseModelHandler):
    """pydantic model that handles serialization internally"""

    def construct(self, message: bytes, schema: Type[BaseModelType]) -> BaseModelType:
        """coverts bytes to a Pydantic model"""
        return schema.parse_raw(message)

    def deconstruct(self, model: BaseModelType) -> bytes:
        """converts a pydantic model to bytes"""
        return model.json().encode("utf-8")


class GenericMessage(BaseModel):
    """The default data model for all profiles.
    Serves as a flexible data model. Accepts extra attributes and provides no type validation.
    """

    class Config:
        extra = Extra.allow


class QueueMessage(BaseModel):
    """message in its raw state off a queue
    message_context: any identifier for a message on a queue.
        used for deleting or markng a message as success after post-processing
    """

    message_context: Any
    message: Any



from volley.serializers.json_serializer import JsonSerialization
from volley.serializers.msgpack_serializer import MsgPackSerialization
from volley.serializers.orjson_serializer import OrJsonSerialization

__all__ = ["JsonSerialization", "MsgPackSerialization", "OrJsonSerialization"]



import json
from typing import Any

from volley.serializers.base import BaseSerialization


class JsonSerialization(BaseSerialization):
    def serialize(self, message: dict[Any, Any]) -> bytes:
        serialized: bytes = json.dumps(message, default=str).encode("utf-8")
        return serialized

    def deserialize(self, message: bytes) -> dict[str, Any]:
        deserialized: dict[str, Any] = json.loads(message)
        return deserialized



from typing import Any

import orjson

from volley.serializers.base import BaseSerialization


class OrJsonSerialization(BaseSerialization):
    def serialize(self, message: dict[Any, Any]) -> bytes:
        serialized: bytes = orjson.dumps(message, option=orjson.OPT_NAIVE_UTC, default=str)
        return serialized

    def deserialize(self, message: bytes) -> dict[str, Any]:
        deserialized: dict[str, Any] = orjson.loads(message)
        return deserialized



# Copyright (c) Shipt, Inc.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass
from typing import Any, List, Optional, Tuple, Union

from prometheus_client import Counter

from volley.logging import logger
from volley.models.base import model_message_handler
from volley.queues import Queue

MESSAGES_PRODUCED = Counter(
    "messages_produced_count", "Messages produced to output destination(s)", ["volley_app", "source", "destination"]
)


@dataclass
class DeliveryReport:
    destination: Optional[str] = None
    asynchronous: bool = False
    status: bool = False


def produce_handler(
    outputs: Union[List[Tuple[str, Any]], List[Tuple[str, Any, dict[str, Any]]]],
    queue_map: dict[str, Queue],
    app_name: str,
    input_name: str,
    message_context: Any,
) -> List[DeliveryReport]:
    """Handles producing messages to output queues

        outputs: List of tuples returned by the Volley app function.
            (queue_name, message_object, optional_runtime_kwargs_for_producer)
        queue_map: mapping of queue_name:Queue object
        app_name: name provided by the application
        input_queue: becomes label in metric counter

    Returns:
        List[DeliveryReport]: DeliveryReport for each destination
    """
    all_delivery_reports: List[DeliveryReport] = []

    for qname, component_msg, *args in outputs:
        try:
            out_queue = queue_map[qname]
        except KeyError as e:
            raise KeyError(f"App not configured for output queue {e}")

        # prepare and validate output message
        if out_queue.data_model is not None and not isinstance(component_msg, out_queue.data_model):
            raise TypeError(f"{out_queue.name=} expected '{out_queue.data_model}' - object is '{type(component_msg)}'")

        try:
            # serialize message
            serialized = model_message_handler(
                data_model=component_msg,
                model_handler=out_queue.model_handler,
                serializer=out_queue.serializer,
            )
            if len(args):
                kwargs: dict[str, Any] = args[0]
            else:
                kwargs = {}
            status = out_queue.producer_con.produce(
                queue_name=out_queue.value, message=serialized, message_context=message_context, **kwargs
            )
            MESSAGES_PRODUCED.labels(volley_app=app_name, source=input_name, destination=qname).inc()
            if not status:
                logger.error("failed producing message to %s" % out_queue.name)
        except Exception:
            logger.exception("failed producing message to %s" % out_queue.name)
            status = False
        all_delivery_reports.append(
            DeliveryReport(
                destination=out_queue.name,
                asynchronous=out_queue.producer_con.callback_delivery,
                status=status,
            )
        )
    return all_delivery_reports


def delivery_success(delivery_reports: List[DeliveryReport]) -> DeliveryReport:
    """determines whether we're going to call on_success or on_fail, or let the producer call it w/ callback
    If there is delivery to an async producer, we'll want to let it handle the callback, but there could
    be synchronous producers that failed as well.

    Producer Cases:
        - single synchronous producer
        - multi synchronous producers
        - single asynchronous producer
        - multi asynchronous producers
        - mixture of synchronous and asynchronous producers

    Order of processing:
        - if a synchronous producer failed - mark it failed
        - if async producer - wait for it
        - else mark it a success
    """

    waiting_async: int = 0
    sync_failures: int = 0
    for report in delivery_reports:
        if report.asynchronous:
            waiting_async += 1
        else:
            sync_failures = sync_failures + int(not report.status)

    if not sync_failures and not waiting_async:
        # happy path. no sync failures and no async producers
        # call on_success synchronously
        return DeliveryReport(status=True, asynchronous=False)
    elif waiting_async and not sync_failures:
        # produced asynchronously without sync fails, return the async delivery report
        # we'll let the Producer call on_success or on_fail
        return DeliveryReport(status=True, asynchronous=True)
    else:
        # had async or synchronous producer(s) failures.
        # call on_fail synchronously
        return DeliveryReport(status=False, asynchronous=False)



# Copyright (c) Shipt, Inc.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Callable, Optional

from volley.data_models import QueueMessage


@dataclass  # type: ignore
class BaseConsumer(ABC):
    """Base class for implementing a consumer"""

    queue_name: str
    host: Optional[str] = None
    config: dict[str, Any] = field(default_factory=dict)

    @abstractmethod
    def consume(self) -> Optional[QueueMessage]:
        """consumes a message from any queue.
        Returns a QueueMessage object on success, or None when there are no messages
        """

    @abstractmethod
    def on_success(self, message_context: Any) -> None:
        """action to take when a message has been successfully consumed.
        For example, delete the message that was consumed.
        """

    @abstractmethod
    def on_fail(self, message_context: Any) -> None:
        """action to perform when serialization, or data validation has failed"""

    @abstractmethod
    def shutdown(self) -> None:
        """perform some action when shutting down the application.
        For example, close a connection or leave a consumer group
        """


@dataclass  # type: ignore
class BaseProducer(ABC):
    """Base class for implementing a producer"""

    queue_name: str
    host: Optional[str] = None
    config: dict[str, Any] = field(default_factory=dict)

    callback_delivery: bool = False
    on_success: Optional[Callable[[Any], None]] = None
    on_fail: Optional[Callable[[Any], None]] = None

    @abstractmethod
    def produce(self, queue_name: str, message: Any, message_context: Optional[Any], **kwargs: Any) -> bool:
        """Publish a message to a queue

        Args:
            queue_name (str): Destination queue name.
            message (Any): The message to publish.
            message_context (Any): Context for the consumed message.
                Often a message id, or a Kafka Message object.
                Used for Producer callbacks to consumer.

        Returns:
            bool: status of the produce operation
        """

    @abstractmethod
    def shutdown(self) -> None:
        """perform some action when shutting down the application"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['volley', 'volley.connectors', 'volley.models', 'volley.serializers']

package_data = \
{'': ['*']}

install_requires = \
['PyRSMQ>=0.4.5,<0.5.0',
 'PyYAML>=5.4.1,<6.0.0',
 'confluent-kafka>=1.7.0,<2.0.0',
 'hiredis>=2.0.0,<3.0.0',
 'msgpack>=1.0.3,<2.0.0',
 'orjson>=3.6.4,<4.0.0',
 'prometheus-client>=0.11.0,<0.12.0',
 'pydantic>=1.8.2,<2.0.0',
 'tenacity>=8.0.1,<9.0.0']

setup_kwargs = {
    'name': 'py-volley',
    'version': '0.0.1a0',
    'description': 'send messages everywhere',
    'long_description': None,
    'author': 'ask-machine-learning',
    'author_email': 'shipt@shipt.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



# Copyright (c) Shipt, Inc.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import asyncio
import builtins
import time
from dataclasses import dataclass, field
from functools import wraps
from typing import Any, Awaitable, Callable, Dict, List, Optional, Tuple, Union

from prometheus_client import Counter, Summary, start_http_server

from volley.concurrency import run_async, run_worker_function
from volley.config import load_yaml
from volley.data_models import QueueMessage
from volley.logging import logger
from volley.models.base import message_model_handler
from volley.profiles import ConnectionType, Profile, construct_profiles
from volley.queues import DLQNotConfiguredError, Queue, construct_queue_map
from volley.transport import DeliveryReport, delivery_success, produce_handler
from volley.util import GracefulKiller

# enables mocking the infinite loop to finite
RUN_ONCE = False


HEARTBEAT = Counter("heartbeats", "Application liveliness")
MESSAGE_CONSUMED = Counter(
    "messages_consumed_count", "Messages consumed from input", ["volley_app", "status"]
)  # success or fail
PROCESS_TIME = Summary("process_time_seconds", "Time spent running a process", ["volley_app", "process_name"])


@dataclass
class Engine:
    """Initializes the Volley application and prepares the main decorator
    Attributes:
        app_name: Name of the application. Added as a label to all logged metrics
        input_queue:
            Name of the input queue.
            Corresponds to the name of one queue defined in queue_config or yaml_config_path
        output_queues: List of queues the application needs to be able to publish to.
            Not required if the application does not produce anywhere!
        dead_letter_queue: Points to the queue in configuration used as the dead letter queue.
        poll_interval_seconds: globally set time to to halt between polls on the consumer
        queue_config: dictionary provided all queue configurations.
            Must provide one of queue_config or yaml_config_path
        yaml_config_path: path to a yaml config file.
            Must provide one of yaml_config_path or queue_config
    Raises:
        NameError: input_queue, dead_letter_queue or output_queues
            reference a queue that does not exist in queue configuration
    Returns:
        Engine: Instance of the engine with a prepared `stream_app` decorator
    """

    input_queue: str
    output_queues: List[str] = field(
        default_factory=list,
    )

    app_name: str = "volley"
    dead_letter_queue: Optional[str] = None

    # set in post_init
    queue_map: Dict[str, Queue] = field(default_factory=dict)
    poll_interval_seconds: float = 1.0
    queue_config: Optional[Dict[str, Any]] = None
    yaml_config_path: str = "./volley_config.yml"
    metrics_port: Optional[int] = 3000

    def __post_init__(self) -> None:
        """Validates configuration and initializes queue configs
        Database connections are initialized within stream_app decorator
        """
        self.killer: GracefulKiller = GracefulKiller()

        if self.output_queues == []:
            logger.warning("No output queues provided")

        # if user provided config, use it
        if self.queue_config:
            cfg = self.queue_config
        else:
            logger.info("loading configuration from %s", self.yaml_config_path)
            cfg = load_yaml(file_path=self.yaml_config_path)["queues"]

        # handle DLQ
        if self.dead_letter_queue is not None:
            # if provided by user, DLQ becomes a producer target
            # flag the queue using the DLQ profile
            if self.dead_letter_queue not in cfg:
                raise KeyError("%s not present in configuration", self.dead_letter_queue)
            else:
                self.output_queues.append(self.dead_letter_queue)
        else:
            logger.warning("DLQ not provided. Application will crash on schema violations")

        # cfg can contain more queues that the app needs
        # filter out queues that are not required by app
        cfg = {k: v for k, v in cfg.items() if k in [self.input_queue] + self.output_queues}

        # validate input_queue, output_queues, and DLQ (optional) are valid configurations
        for q in [self.input_queue] + self.output_queues:
            if q not in cfg:
                raise KeyError(f"Queue '{q}' not found in configuration")

        # tag queues with type (how app intends to use it)
        cfg[self.input_queue]["connection_type"] = ConnectionType.CONSUMER
        for qname in self.output_queues:
            cfg[qname]["connection_type"] = ConnectionType.PRODUCER

        # load profiles
        profiles: Dict[str, Profile] = construct_profiles(cfg)
        # create queue_map from profiles
        self.queue_map = construct_queue_map(profiles, cfg)

        logger.info("Queues initialized: %s", list(self.queue_map.keys()))

    def stream_app(  # noqa: C901
        self,
        func: Callable[
            [Any], Union[Awaitable[Any], List[Tuple[str, Any]], List[Tuple[str, Any, dict[str, Any]]], bool]
        ],
    ) -> Callable[..., None]:
        """Main decorator for applications"""

        is_coroutine = asyncio.iscoroutinefunction(func)

        @run_async
        @wraps(func)
        async def run_component() -> None:
            if self.metrics_port is not None:
                start_http_server(port=self.metrics_port)
            # the component function is passed in as `func`
            # first setup the connections to the input and outputs queues that the component will need
            # we only want to set these up once, before the component is invoked

            # initialize connections to each queue, and schemas
            self.queue_map[self.input_queue].connect(con_type=ConnectionType.CONSUMER)
            # we only want to connect to queues passed in to Engine()
            # there can be more queues than we need defined in the configuration yaml
            for out_q_name in self.output_queues:
                self.queue_map[out_q_name].connect(con_type=ConnectionType.PRODUCER)

            # queue connections were setup above. now we can start to interact with the queues
            # alias for input connection readability
            input_con: Queue = self.queue_map[self.input_queue]

            # if asynchronous producer, give the consumer's "on_success" method to the producer
            for qname in self.output_queues:
                producer_con = self.queue_map[qname].producer_con
                if producer_con.callback_delivery:
                    producer_con.init_callbacks(consumer=self.queue_map[self.input_queue].consumer_con)

            logger.info("Starting Volley application: %s -- %s", self.app_name, not self.killer.kill_now)
            while not self.killer.kill_now:
                HEARTBEAT.inc()
                _start_time = time.time()
                # read message off the specified queue
                in_message: Optional[QueueMessage] = input_con.consumer_con.consume()
                if in_message is None:
                    # if no messages, handle poll interval
                    # TODO: this should be dynamic with some sort of backoff
                    logger.info("No messages - sleeping POLL_INTERVAL=%s", self.poll_interval_seconds)
                    time.sleep(self.poll_interval_seconds)
                    continue

                # typing for producing
                outputs: Union[List[Tuple[str, Any]], List[Tuple[str, Any, dict[str, Any]]], bool] = False

                data_model, consume_status = message_model_handler(
                    message=in_message.message,
                    schema=input_con.data_model,
                    model_handler=input_con.model_handler,
                    serializer=input_con.serializer,
                )

                if consume_status:
                    # happy path
                    MESSAGE_CONSUMED.labels(volley_app=self.app_name, status="success").inc()
                elif self.dead_letter_queue is not None and self.dead_letter_queue in self.queue_map:
                    outputs = [(self.dead_letter_queue, data_model)]
                    MESSAGE_CONSUMED.labels(volley_app=self.app_name, status="fail").inc()
                else:
                    # things have gone wrong w/ the message and no DLQ configured
                    MESSAGE_CONSUMED.labels(volley_app=self.app_name, status="fail").inc()
                    raise DLQNotConfiguredError(f"Deserializing {in_message.message} failed")

                # component processing
                if not outputs:
                    # this is happy path
                    # if outputs have been assigned it means this message is destined for a DLQ
                    _start_main = time.time()
                    outputs = await run_worker_function(func, message=data_model, is_coroutine=is_coroutine)
                    _fun_duration = time.time() - _start_main
                    PROCESS_TIME.labels(volley_app=self.app_name, process_name="component").observe(_fun_duration)

                delivery_report: DeliveryReport
                if isinstance(outputs, builtins.bool):
                    # if func returns a bool, its just a bool and nothing more
                    # there is no "producer" in this model.
                    # Just mark the consumed message as either success or fail
                    delivery_report = DeliveryReport(status=outputs, asynchronous=False)

                else:
                    delivery_reports: list[DeliveryReport] = produce_handler(
                        outputs=outputs,
                        queue_map=self.queue_map,
                        app_name=self.app_name,
                        input_name=input_con.name,
                        message_context=in_message.message_context,
                    )
                    delivery_report = delivery_success(delivery_reports)
                if delivery_report.status is True and not delivery_report.asynchronous:
                    # asynchronous delivery reports are handled within the Producer's callback
                    # synchrnous delivery reports are handled here
                    input_con.consumer_con.on_success(
                        message_context=in_message.message_context,
                    )
                elif not delivery_report.asynchronous:
                    input_con.consumer_con.on_fail(
                        message_context=in_message.message_context,
                    )
                _duration = time.time() - _start_time
                PROCESS_TIME.labels(volley_app=self.app_name, process_name="cycle").observe(_duration)

                if RUN_ONCE:
                    # for testing purposes only - mock RUN_ONCE
                    break

            self.shutdown()
            logger.info("Shutdown %s complete", self.app_name)

        # used for unit testing as a means to access the wrapped component without the decorator
        run_component.__wrapped__ = func  # type: ignore

        return run_component

    def shutdown(self) -> None:
        """graceful shutdown of all queue connections"""
        logger.info("Shutting down %s, %s", self.app_name, self.queue_map[self.input_queue].value)
        for q_name in self.output_queues:
            out_queue = self.queue_map[q_name]
            logger.info("Shutting down %s, %s", self.app_name, out_queue.value)
            out_queue.producer_con.shutdown()
            logger.info("%s, %s shutdown complete", self.app_name, q_name)
        self.queue_map[self.input_queue].consumer_con.shutdown()



from volley.models.pydantic_model import (
    GenericMessage,
    PydanticModelHandler,
    PydanticParserModelHandler,
)

__all__ = ["GenericMessage", "PydanticModelHandler", "PydanticParserModelHandler"]



# Copyright (c) Shipt, Inc.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from volley.config import import_module_from_string
from volley.models.base import BaseModelHandler
from volley.profiles import ConnectionType, Profile
from volley.serializers.base import BaseSerialization


class DLQNotConfiguredError(Exception):
    """raised when message processes fails and no DLQ configured"""

    pass


@dataclass
class Queue:
    """a Queue object represents everything we need to know about a queue"""

    # alias for the queue
    name: str
    # system name for the queue. for example, some really long kafka topic name
    value: str
    profile: Profile

    # instantiated post-init
    data_model: Optional[Any] = field(init=False, default=None)
    model_handler: Optional[BaseModelHandler] = field(init=False, default=None)
    serializer: Optional[BaseSerialization] = field(init=False, default=None)

    # initialized queue connections
    # these are initialized by calling connect()
    # consumer_con: Consumer
    # producer_con: Producer

    # optional configurations to pass through to connectors
    pass_through_config: dict[str, str] = field(default_factory=dict)

    def __post_init__(self) -> None:
        """Load modules provided in Profile"""
        # data model is assigned by not instantiated
        if self.profile.data_model is not None:
            self.data_model = import_module_from_string(self.profile.data_model)

        # model_handler and serializer are instantiated on init
        if self.profile.model_handler is not None:
            self.model_handler = import_module_from_string(self.profile.model_handler)()
        if self.profile.serializer is not None:
            self.serializer = import_module_from_string(self.profile.serializer)()

    def connect(self, con_type: ConnectionType) -> None:
        """instantiate the connector class"""
        if con_type == ConnectionType.CONSUMER:
            if self.profile.consumer is None:
                raise ValueError("Must provide a consumer connector")
            _class = import_module_from_string(self.profile.consumer)
            self.consumer_con = _class(queue_name=self.value, config=self.pass_through_config.copy())
        elif con_type == ConnectionType.PRODUCER:
            if self.profile.producer is None:
                raise ValueError("Must provide a producer connector")
            _class = import_module_from_string(self.profile.producer)
            self.producer_con = _class(queue_name=self.value, config=self.pass_through_config.copy())
        else:
            raise TypeError(f"{con_type=} is not valid")


def construct_queue_map(profiles: Dict[str, Profile], configs: dict[str, dict[str, str]]) -> Dict[str, Queue]:
    """Constructs a mapping of queue_name: Queue for each requested queue"""

    queue_map: Dict[str, Queue] = {}
    for qname, profile in profiles.items():
        q_config = configs[qname]
        cfg: Any = q_config.get("config", {})
        if (value := q_config.get("value")) is None:
            raise KeyError(f"Must provide `value` in configuration for `{qname}`")
        queue = Queue(
            name=qname,
            value=value,
            profile=profile,
            pass_through_config=cfg,
        )
        queue_map[qname] = queue

    return queue_map
